functor TypHelper(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    sharing TEnv.Typ = Typ
	        and Typ.Globals = Globals
	        and TEnv.Globals = Globals
	        and Typ.Id = Globals.Id
	    ) : TYPHELPER
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TEnv = TEnv
open TEnv
open Typ
open Globals
open Pp

datatype pretyp = 
	    PRETVAR of Id.T
	  | PREARROW of pretyp * pretyp
	  | PREALL of Id.T * pretyp * pretyp
	  | PRESOME of Id.T * pretyp * pretyp
	  | PREMEET of pretyp list
	  | PREOBJ of Id.T * pretyp
	  | PREMINUS of pretyp * (Id.T list)

fun debruijnify te (PRETVAR i) =
      tvar (TEnv.index te i)
  | debruijnify te (PREARROW (pt1,pt2)) =
      arrow (debruijnify te pt1) (debruijnify te pt2)
  | debruijnify te (PREALL (i,pt1,pt2)) =
      all i (debruijnify te pt1) (debruijnify (TEnv.push_bound te i NS) pt2)
  | debruijnify te (PRESOME (i,pt1,pt2)) =
      some i (debruijnify te pt1) (debruijnify (TEnv.push_bound te i NS) pt2)
  | debruijnify te (PREMEET pts) =
      meet (map (fn pt => debruijnify te pt) pts)
  | debruijnify te (PREOBJ (n,pt1)) =
      obj n (debruijnify te pt1) 
  | debruijnify te (PREMINUS (pt1,ids)) =
      minus (debruijnify te pt1) ids


(* ----------------------------------------------------------------------------- *)

val PRT_NS_AS_T = createflag ["prtT"] true

fun prt_ns () = 
  if !PRT_NS_AS_T then write "T" else write "ns" 

fun prt_nested te t =
  let fun p te t =
    case explode(t) of
          EXPTVAR i =>
             (write (Id.tostr (lookup_name te i)))
        | EXPARROW (t1,t2) =>
              (write "(";
               p te t1;
               write "->";
               p te t2;
               write ")")
        | EXPALL (i,t1,t2) => let
      	       val te' = push_bound te i t1
      	       in write "(All ";
		  write (Id.tostr (lookup_name te' 0));
		  if not (eq t1 NS)
		    then (write "<";
			  p te t1)
		    else ();
		  write ". ";
		  p te' t2;
		  write ")"
	        end
        | EXPSOME (i,t1,t2) => let
      	       val te' = push_bound te i t1
      	       in write "(Some ";
		  write (Id.tostr (lookup_name te' 0));
		  if not (eq t1 NS)
		    then (write "<";
			  p te t1)
		    else ();
		  write ". ";
		  p te' t2;
		  write ")"
	        end
        | EXPMEET ts =>
              if null ts
                then prt_ns()
                else (write "/\\[";
		      plist te ts;
		      write "]")
        | EXPOBJ (n,t) =>
              (write (Id.tostr n);
               write ":";
               p te t)
        | EXPMINUS (t,ids) =>
              (p te t;
	       write "\\";
	       case ids of
		   [i] => write (Id.tostr i)
		 | _   => (write "{";
			   mapunit_tuple 
			      (fn i=> write (Id.tostr i)) (fn()=> write",") 
			      ids;
			   write "}")
               )
      and plist te [] = 
            ()
        | plist te [t] = 
            p te t
        | plist te (hd::tl) = 
            (p te hd; write ","; plist te tl)
  in p te t
  end


val PRINTLOWLEVEL = createflag ["printlowlevel"] false

val printmeetsinfix = createflag ["printmeetsinfix", "infixmeets"] false
val printtvarnumbers = createflag ["printtvarnumbers", "tvarnumbers"] false
	 
(* Precedence codes:
	outerprec -- no parens needed
	meetprec -- parens needed for /\
	objprec -- parens needed for /\, i:
	allprec  -- needed for /\, All, i:
	arrowprec -- needed for /\, All, ->, i:
*)
val outerprec = 0
val meetprec = 1
val objprec = 2
val allprec = 3
val arrowprec = 4

(* $$$ check that precedence of \{...} is the same on outut as on input *)
fun prt_ind te outer t =
  let fun pparen prec minprec paren = 
        if prec >= minprec then write paren else ()
      fun ptcon prec outer s =
	if outer
	  then (write " "; ubreak (~1 - (size s)); write s; write " ")
	  else (ubreak (0 - (size s)); write s)
      fun p te prec outer t = 
	case explode(t) of
	      EXPTVAR i => 
              (write (Id.tostr (lookup_name te i));
               if (!printtvarnumbers)
                 then write ("^"^(makestring i))
                 else ())
            | EXPARROW (t1,t2) => 
              (pparen prec arrowprec "(";
               setb();
               p te arrowprec false t1;
               ptcon prec outer "->";
               p te allprec outer t2;
               endb();
               pparen prec arrowprec ")")
            | EXPALL (i,t1,t2) =>
              (pparen prec allprec "(";
               setb();
               write "All ";
               write (Id.tostr i);
               if (eq t1 NS)
                 then ()
                 else (write "<";
		       p te arrowprec false t1);
               write ". ";
               ubreak 2;
               p (push_bound te i t1) objprec outer t2;
               endb();
               pparen prec allprec ")")
            | EXPSOME (i,t1,t2) =>
              (pparen prec allprec "(";
               setb();
               write "Some ";
               write (Id.tostr i);
               if (eq t1 NS)
                 then ()
                 else (write "<";
		       p te arrowprec false t1);
               write ". ";
               ubreak 2;
               p (push_bound te i t1) objprec outer t2;
               endb();
               pparen prec allprec ")")
            | EXPMEET ts =>
               if (!printmeetsinfix) 
                 then pmeetlist te prec outer ts
               	 else pmeetinfix te prec outer ts
            | EXPOBJ (n,t) =>
               (write (Id.tostr n);
                write ":";
                p te meetprec false t)
            | EXPMINUS (t,ids) =>
               (p te arrowprec false t;
	        write "\\";
	        case ids of
		    [i] => write (Id.tostr i)
		  | _   => (write "{";
			    mapunit_tuple 
			       (fn i=> write (Id.tostr i)) (fn()=> write",") 
			       ids;
			    write "}")
               )
      and pmeetlist te prec outer [] =
               prt_ns()
        | pmeetlist te prec outer (ts) =
              (write "/\\[";
               plist te outer ts;
               write "]")
      and pmeetinfix te prec outer [] =
               prt_ns()
        | pmeetinfix te prec outer [t] =
              (write "/\\[";
               p te 0 false t;
               write "]")
        | pmeetinfix te prec outer ts =
              (pparen prec meetprec "(";
               setb();
               (* Seems to crash the prettyprinter... $$$ *)
               (* expbreak true "   "; *)
               pmeets te prec outer ts;
               endb();
               pparen prec meetprec ")")
      and pmeets te _ outer [] = ()
        | pmeets te _ outer [t] = 
              p te objprec false t
        | pmeets te prec outer (t::ts) = 
             (p te objprec false t;
              ptcon prec outer "/\\";
              pmeets te prec outer ts)
      and plist' te outer [] = 
            ()
        | plist' te outer [t] = 
            p te 0 false t
        | plist' te outer (hd::tl) = 
            (p te 0 false hd;
             write ","; 
             if outer then write " " else ();
             break false 0;
             plist te outer tl)
      and plist te outer ts =
            (setb();
             plist' te outer ts;
             endb())
  in 
    p te outerprec outer t
  end

fun prt te t =
  if !PRINTLOWLEVEL
    then prt_exactly t
    else prt_ind te false t

fun prt_outer te t =
  if !PRINTLOWLEVEL
    then prt_exactly t
    else prt_ind te true t

val short_tenvs = createflag ["shorttenvs"] true

fun prt_tenv te = let
  fun p_one n = 
        (write (Id.tostr (lookup_name te n));
         if !PRINTLOWLEVEL
           then bare_lookup te n
		  (fn x => (write "<"; prt_exactly x))
		  (fn x => (write ":"; prt_exactly x))
		  (fn x => (write "=="; prt_exactly x))
           else lookup_and_relocate te n
		  (fn x => (write "<"; prt te x))
		  (fn x => (write ":"; prt te x))
		  (fn x => (write "=="; prt te x))  )
  fun p 0 = p_one 0
    | p n = (p_one n;
             write ", ";
             uubreak 0;
             p (n-1))
  val tesize = domsize te
  in (* write "{"; *)
     setb();
     if !short_tenvs
       then if tesize > 0 then (write "..."; p_one 0) else ()
       else (p (tesize-1));
     endb()
     (* ;write "}" *)
  end


end

