functor TypHelper(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    sharing TEnv.Typ = Typ
	        and Typ.Globals = Globals
	        and TEnv.Globals = Globals
	        and Typ.Id = Globals.Id
	    ) : TYPHELPER
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TEnv = TEnv
open TEnv
open Typ
open Globals
open Pp

datatype PreKind = 
	    PRESTAR
	  | PREKARROW of PreKind * PreKind

and PreTyp = 
	    PRETVAR of Id.T
	  | PREABST of Id.T * PreKind * PreTyp
	  | PREAPPT of PreTyp * PreTyp
	  | PREARROW of PreTyp * PreTyp
	  | PREALL of Id.T * PreTyp * PreTyp
	  | PRESOME of Id.T * PreTyp * PreTyp
	  | PRETOP of PreKind
	  | PRERCD of (Id.T * PreTyp) list

fun debruijnifyk te PRESTAR = star
  | debruijnifyk te (PREKARROW(pk1,pk2)) = 
      karrow (debruijnifyk te pk1) (debruijnifyk te pk2)

and debruijnify te (PRETVAR i) =
      tvar (TEnv.index te i)
  | debruijnify te (PREABST (n,pk1,pt2)) =
      abst n (debruijnifyk te pk1) 
      	     (debruijnify (TEnv.push_bound te n dummytype) pt2)
  | debruijnify te (PREAPPT (pt1,pt2)) =
      appt (debruijnify te pt1) (debruijnify te pt2)
  | debruijnify te (PREARROW (pt1,pt2)) =
      arrow (debruijnify te pt1) (debruijnify te pt2)
  | debruijnify te (PREALL (i,pt1,pt2)) =
      all i (debruijnify te pt1) 
      	    (debruijnify (TEnv.push_bound te i dummytype) pt2)
  | debruijnify te (PRESOME (i,pt1,pt2)) =
      some i (debruijnify te pt1) 
      	     (debruijnify (TEnv.push_bound te i dummytype) pt2)
  | debruijnify te (PRETOP(prek)) =
      top (debruijnifyk te prek)
  | debruijnify te (PRERCD(l)) =
      rcd (map (fn (n,pt) => (n, debruijnify te pt)) l)

fun keq te k1 k2 = let
  fun e (STAR) (STAR) = true
    | e (KARROW(k11,k12)) (KARROW(k21,k22)) = 
      (keq te k11 k21) andalso (keq te k12 k22)
    | e _ _ = false
  in e k1 k2
  end

and eq te s t = let
  fun e (TVAR(i)) (TVAR(i')) = 
        (i = i')
        orelse
        (lookup_bound_or_abbrev te i
          (fn _ => lookup_bound_or_abbrev te i'
          	     (fn _ => false)
          	     (fn abb => eq te s abb))
          (fn abb => eq te abb t))
    | e (ABST(a,ks1,s2)) (ABST(_,kt1,t2)) = 
          (keq te ks1 kt1)
          andalso
          (eq (push_bound te a dummytype) s2 t2)
    | e (APPT(sf,sa)) (APPT(tf,ta)) = 
          (eq te sf tf) 
          andalso
          (eq te sa ta)
    | e (ARROW(s1,s2)) (ARROW(t1,t2)) = 
          (eq te s1 t1) 
          andalso
          (eq te s2 t2)
    | e (ALL(a,s1,s2)) (ALL(_,t1,t2)) = 
          (eq te s1 t1)
          andalso
          (eq (push_bound te a dummytype) s2 t2)
    | e (SOM(a,s1,s2)) (SOM(_,t1,t2)) = 
          (eq te s1 t1)
          andalso
          (eq (push_bound te a dummytype) s2 t2)
    | e (TOP(ks)) (TOP(kt)) = 
          (keq te ks kt)
    | e (RCD(ls)) (RCD(lt)) = 
          (forall (fn (i,s) => case Id.lookup_field i lt of
                                   NONE => false
                                 | SOME(t) => eq te s t)
                  ls)
          andalso
          (forall (fn (i,t) => case Id.lookup_field i ls of
                                   NONE => false
                                 | SOME(_) => true)
                  lt)
    | e _ _ = false
  in e s t
  end
  

(* ----------------------------------------------------------------------- *)

val typeabbrevs = createflag ["typeabbrevs"] true

datatype AllSome = PALL | PSOME

datatype KP = 
    PSTAR 
  | PKARROW of KP list

datatype P = 
    PTVAR of string
  | PABST of ((string * KP) list) * P
  | PAPPT of P list
  | PARROW of P list
  | PALLSOME of ((AllSome * string * P) list) * P
  | PTOP of KP
  | PRCD of (string * P) list

fun kprintify k = let
  fun kp STAR = PSTAR
    | kp (ek as KARROW(_)) = PKARROW (coll_karrow k)
  in kp k
  end

and coll_karrow k = let
  fun c STAR = [PSTAR]
    | c (KARROW(k1,k2)) = (kprintify k1) :: (coll_karrow k2)
  in c k
  end

fun printify te t = let
  fun p (TVAR(i)) = PTVAR(Id.tostr (TEnv.lookup_name te i))
    | p (ABST(_)) = PABST(coll_abst te t)
    | p (APPT(t1,t2)) = PAPPT(coll_appt te t)
    | p (ARROW(_)) = PARROW(coll_arrow te t)
    | p (ALL(_)) = PALLSOME(coll_allsome te t)
    | p (SOM(_)) = PALLSOME(coll_allsome te t)
    | p (TOP(k)) = PTOP(kprintify k)
    | p (RCD(fields)) = PRCD(map (fn(l,t)=> (Id.tostr l, printify te t)) fields)
  in p t
  end

and coll_abst te t = let
  fun c (ABST(a,k,t)) = let
      val (prefix,body) = coll_abst (TEnv.push_bound te a dummytype) t
      in ((Id.tostr a, kprintify k)::prefix,body) end 
    | c _ = ([], printify te t)
  in c t
  end

and coll_appt te t = let
  fun c (APPT(t1,t2)) = (coll_appt te t1) @ [printify te t2]
    | c _ = [printify te t]
  in c t
  end

and coll_arrow te t = let
  fun c (ARROW(t1,t2)) = (printify te t1) :: (coll_arrow te t2)
    | c _ = [printify te t]
  in c t
  end

and coll_allsome te t = let
  fun c (ALL(a,t1,t2)) = let
      val (prefix,body) = coll_allsome (TEnv.push_binding te a dummytype) t2
      in ((PALL, Id.tostr a, printify te t1)::prefix,body)
      end
    | c (SOM(a,t1,t2)) = let
      val (prefix,body) = coll_allsome (TEnv.push_binding te a dummytype) t2
      in ((PSOME, Id.tostr a, printify te t1)::prefix,body)
      end
    | c _ = ([], printify te t)
  in c t
  end

fun kprt_formatted k = let
  fun p PSTAR inner = 
      write "*"
    | p (PKARROW(ks)) inner = 
      (if inner then write "(" else ();
       mapunit_tuple 
          (fn(ek)=> p ek true)
          (fn()=> write "->")
          ks;
       if inner then write ")" else ())
  in p k false
  end

val indent = 2

val INDENT_FIRST_ARROW = createflag ["indent_first_arrow"] true
val UNITE_RCDS = createflag ["unite_rcds"] false

fun prt_formatted t nested outer = let
  fun p (PTVAR(n)) = 
      write n
    | p (PABST(bnd,body)) = 
      (mapunit_tuple
         (fn (a,k) => (write "Fun("; write a; 
                       if !typeabbrevs andalso k=PSTAR
                         then ()
                         else (write ":"; kprt_formatted k); 
                       write ")"; if outer then write " " else ()))
         (fn()=> ubreak 0)
         bnd;
       ubreak indent;
       prt_formatted body false outer)
    | p (PAPPT([])) = 
       raise CantHappen
    | p (PAPPT([t])) = 
       raise CantHappen
(* This is a good idea, but it needs to be done more carefully: 
    | p (PAPPT([PTVAR(n),t2])) = 
      (* Special syntax for applying a type variable to one argument: A(N) *)
      (write n; write "(";
       setb();
       prt_formatted t2 false false;
       write ")"; endb()) *)
    | p (PAPPT(first::rest)) = 
      (prt_formatted first false outer;
       ubreak indent;
       setb();
       mapunit 
         (fn (t) => (ubreak 0; write " "; prt_formatted t true false))
         rest;
       endb())
    | p (PARROW(elts)) = 
      (if !INDENT_FIRST_ARROW
         then if outer 
                   then expbreak true "   "
                   else expbreak true "  "
         else ();
       mapunit_tuple
         (fn (t) => prt_formatted t true false)
         (fn()=> if outer 
                   then (ubreak ~1; write " -> ")
                   else (ubreak 0; write "->"))
         elts)
    | p (PALLSOME(bnd,body)) = 
      (mapunit_tuple
         (fn (tag,a,t1) => (case tag of PALL => write "All" | PSOME => write "Some";
                            write "("; write a; 
                            if !typeabbrevs andalso t1=PTOP(PSTAR)
                              then ()
                              else (write "<"; prt_formatted t1 false false);
                            write ")"; 
                            if outer then write " " else ()))
         (fn()=> ubreak 0)
         bnd;
       ubreak indent;
       prt_formatted body false outer)
    | p (PTOP(k)) = 
      (if !typeabbrevs andalso k=PSTAR
         then write "Top"
         else (write "Top(";
               kprt_formatted k;
               write ")"))
    | p (PRCD(fields)) =
      (write "{|";
       mapunit_tuple
         (fn(l,t)=> (write l; write ":"; if outer then write " " else (); 
                     prt_formatted t false false))
         (fn() => (write ", "; if outer then write " " else (); 
                   if !UNITE_RCDS then ubreak 2 else uubreak 2))
         fields;         
       write "|}")
  fun shownesting _ false _ = ()
(*  | shownesting (PAPPT([PTVAR(_),_])) _ _ = () *)
    | shownesting (PRCD(_)) _ _ = ()
    | shownesting (PTVAR(_)) _ _ = ()
    | shownesting (PTOP(_)) _ _ = ()
    | shownesting _ _ n = write n
  in 
    shownesting t nested "(";
    setb();
    p t;
    endb();
    shownesting t nested ")"
  end

(* ----------------------------------------------------------------------- *)

val PRINTLOWLEVEL = createflag ["printlowlevel"] false

fun prt_nested te t = prt_exactly t

(* High-level / low-level printing switches *)
fun kprt te k = if !PRINTLOWLEVEL then kprt_exactly k else kprt_formatted (kprintify k)
fun prt te t = if !PRINTLOWLEVEL then prt_exactly t else prt_formatted (printify te t) false false
fun prt_outer te t = if !PRINTLOWLEVEL then prt_exactly t else prt_formatted (printify te t) false true

(* ----------------------------------------------------------------------- *)

val short_tenvs = createflag ["shorttenvs"] true

fun prt_tenv te = let
  fun p_one n = 
        (write (Id.tostr (lookup_name te n));
         if !PRINTLOWLEVEL
           then bare_lookup te n
		  (fn x => (write "<"; prt_exactly x))
		  (fn x => (write ":"; prt_exactly x))
		  (fn x => (write "=="; prt_exactly x))
           else lookup te n
		  (fn x => (write "<"; prt te x))
		  (fn x => (write ":"; prt te x))
		  (fn x => (write "=="; prt te x))  )
  fun p 0 = p_one 0
    | p n = (p_one n;
             write ", ";
             uubreak 0;
             p (n-1))
  val tesize = domsize te
  in (* write "{"; *)
     setb();
     if !short_tenvs
       then if tesize > 0 then (write "..."; p_one 0) else ()
       else (p (tesize-1));
     endb()
     (* ;write "}" *)
  end

(* ----------------------------------------------------------------------- *)

val DEBUG = createflag ["shownorm"] false

fun normalize te t = let
  fun n (TVAR(i)) = lookup_bound_or_abbrev te i
  			 (fn _ => t)
  			 (fn teOFt => normalize te teOFt)
    | n (ABST(a,k1,t2)) = let
        in abst a k1 (normalize (push_bound te a (top k1)) t2)
        end
    | n (APPT(tf,ta)) = let
        val tfn = (normalize te tf)
        val tan = (normalize te ta)
        in case tfn of
             ABST(n,tf1,tf2) => normalize te (tsubst_top tan tf2)
           | _ => appt tfn tan
        end
    | n (ARROW(t1,t2)) = 
        arrow (normalize te t1) (normalize te t2)
    | n (ALL(a,t1,t2)) = let
        val t1n = (normalize te t1)
        in all a t1n (normalize (push_bound te a t1n) t2)
        end
    | n (SOM(a,t1,t2)) = let
        val t1n = (normalize te t1)
        in some a t1n (normalize (push_bound te a t1n) t2)
        end
    | n (TOP(k)) = top(k)
    | n (RCD(l)) = rcd (map (fn (n,t) => (n, normalize te t)) l)
  in 
     wrap DEBUG "normalize"
       (fn()=> n t)
       (fn()=> prt te t)
       (fn tn => prt te tn)
  end

fun whn te t = let
  fun n (TVAR(i)) = lookup_bound_or_abbrev te i
  			 (fn _ => t)
  			 (fn teOFt => whn te teOFt)
    | n (ABST(a,k1,t2)) = t
    | n (APPT(tf,ta)) = let
        val tfN = whn te tf
        in case tfN of
             ABST(n,tf1,tf2) => whn te (tsubst_top ta tf2)
           | _ => appt tfN ta
        end
    | n (ARROW(t1,t2)) = t
    | n (ALL(a,t1,t2)) = t
    | n (SOM(a,t1,t2)) = t
    | n (TOP(k)) = t
    | n (RCD(l)) = t
  in 
     wrap DEBUG "whn"
       (fn()=> n t)
       (fn()=> prt te t)
       (fn tn => prt te tn)
  end

val DEBUGBETAEQ = createflag ["betaeq"] false
val DEBUGBETAEQENVS = createflag ["betaeqenvs"] false

fun betaeq te s t = let
  fun b (TVAR(i)) (TVAR(i')) = 
      (i = i')
    | b (ABST(a,k11,t12)) (ABST(_,k21,t22)) = 
      (keq te k11 k21)
      andalso
      (betaeq (push_bound te a dummytype) t12 t22)
    | b (APPT(s1,s2)) (APPT(t1,t2)) =
      (betaeq te s1 t1)
      andalso
      (betaeq te s2 t2)
    | b (ARROW(s1,s2)) (ARROW(t1,t2)) =
      (betaeq te s1 t1)
      andalso
      (betaeq te s2 t2)
    | b (ALL(a,s1,s2)) (ALL(_,t1,t2)) =
      (betaeq te s1 t1)
      andalso
      (betaeq (push_bound te a dummytype) s2 t2)
    | b (SOM(a,s1,s2)) (SOM(_,t1,t2)) =
      (betaeq te s1 t1)
      andalso
      (betaeq (push_bound te a dummytype) s2 t2)
    | b (TOP(ks)) (TOP(kt)) =
      keq te ks kt
    | b (RCD(ls)) (RCD(lt)) = 
          (forall (fn (i,s) => case Id.lookup_field i lt of
                                   NONE => false
                                 | SOME(t) => betaeq te s t)
                  ls)
          andalso
          (forall (fn (i,t) => case Id.lookup_field i ls of
                                   NONE => false
                                 | SOME(_) => true)
                  lt)
    | b _ _ = false
  in 
     wrap DEBUGBETAEQ "betaeq"
       (fn()=> b (whn te s) (whn te t))
       (fn()=> (setb(); prt te s; ubreak ~2; 
                write " =beta "; prt te t;
                if not (!DEBUGBETAEQENVS) then () else
                (write "   "; ubreak 0; write "in ";
                prt_tenv te);
                endb()))
       (fn b=> write (if b then "Yes" else "No"))
  end

end

